import os
from django.test.utils import setup_test_environment
import django
from django.core.management import call_command

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "tests.settings")
setup_test_environment()
django.setup()

if getattr(django.conf.settings, "VARIANT") == "spatialite":
    # See https://code.djangoproject.com/ticket/32935
    from django.db import connection

    with connection.cursor() as cursor:
        cursor.execute("SELECT InitSpatialMetaData(1);")
    connection.close()

call_command("makemigrations", interactive=False)
call_command("migrate", interactive=False)



# Generated by Django 4.0.3 on 2022-03-31 04:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import tests.rest_app.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='CharFieldModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('required_field', models.CharField(max_length=20)),
                ('nullable_field', models.CharField(max_length=20, null=True)),
                ('blankable_field', models.CharField(blank=True, max_length=20)),
                ('nullableblankable_field', models.CharField(blank=True, max_length=20, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ChoiceModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Enter Name', max_length=10)),
                ('choice', models.CharField(choices=[('one', 'Choice One'), ('two', 'Choice Two'), ('three', 'Choice Three')], help_text='Pick One', max_length=10)),
                ('grouped_choice', models.CharField(choices=[('Group 1', (('one', 'Choice One'), ('two', 'Choice Two'), ('three', 'Choice Three')))], help_text='Pick One', max_length=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='DateModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('date', models.DateTimeField()),
                ('empty_date', models.DateTimeField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='ExpensiveModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=20)),
                ('expensive', tests.rest_app.models.ExpensiveField(cost=2, max_length=255)),
                ('more_expensive', tests.rest_app.models.ExpensiveField(blank=True, cost=5, max_length=255, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FieldsetModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('title', models.CharField(max_length=20)),
                ('address', models.CharField(max_length=255)),
                ('city', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FileModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('file', models.FileField(upload_to='files')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ImageModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('image', models.ImageField(upload_to='files')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ItemType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('active', models.NullBooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Parent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RootModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.SlugField()),
                ('description', models.TextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SlugModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.SlugField()),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TranslatedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='translated model name')),
            ],
        ),
        migrations.CreateModel(
            name='UserManagedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='SlugRefParent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('ref', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='rest_app.slugmodel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SlugRefChild',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='rest_app.slugrefparent')),
            ],
        ),
        migrations.CreateModel(
            name='OneToOneModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('root', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='rest_app.rootmodel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='rest_app.itemtype')),
            ],
            options={
                'ordering': ('type', 'name'),
            },
        ),
        migrations.CreateModel(
            name='ForeignKeyModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('root', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='rest_app.rootmodel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ExtraModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('alt_root', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='extramodel_set', to='rest_app.rootmodel')),
                ('root', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='extramodels', to='rest_app.rootmodel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Child',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='children', to='rest_app.parent')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import unittest
from .base import APITestCase
import json
from django.conf import settings


class ConfigTestCase(APITestCase):
    def get_config(self, page_name):
        response = self.client.get("/config.json")
        result = json.loads(response.content.decode("utf-8"))
        self.assertIn("pages", result)
        self.assertIn(page_name, result["pages"])
        return result["pages"][page_name]

    def get_field(self, page_config, field_name, allow_none=False):
        self.assertIn("form", page_config)
        for field in page_config["form"]:
            if field["name"] == field_name:
                return field
        if not allow_none:
            self.fail("Could not find %s" % field_name)

    # Test existence and content of config.json
    def test_rest_config_json(self):
        response = self.client.get("/config.json")
        result = json.loads(response.content.decode("utf-8"))
        self.assertIn("pages", result)

        # Extra config
        self.assertIn("debug", result)

    def test_rest_config_meta(self):
        conf = self.get_config("item")
        self.assertEqual(
            {
                "name": "item",
                "url": "items",
                "list": True,
                "cache": "none",
                "form": conf["form"],
                "label_template": "{{name}}",
                "verbose_name": "item",
                "verbose_name_plural": "items",
                "ordering": ["type", "name"],
            },
            conf,
        )

    def test_rest_config_json_fields(self):
        self.assertEqual(
            [
                {
                    "name": "name",
                    "label": "Name",
                    "type": "string",
                    "wq:length": 10,
                    "bind": {"required": True},
                },
                {
                    "name": "date",
                    "label": "Date",
                    "type": "dateTime",
                    "bind": {"required": True},
                },
                {
                    "name": "empty_date",
                    "label": "Empty Date",
                    "type": "dateTime",
                },
            ],
            self.get_config("datemodel")["form"],
        )

    def test_rest_config_json_choices(self):
        conf = self.get_config("choicemodel")
        self.assertEqual(
            [
                {
                    "name": "name",
                    "label": "Name",
                    "hint": "Enter Name",
                    "type": "string",
                    "wq:length": 10,
                    "bind": {"required": True},
                },
                {
                    "name": "choice",
                    "label": "Choice",
                    "hint": "Pick One",
                    "type": "select one",
                    "bind": {"required": True},
                    "choices": [
                        {
                            "name": "one",
                            "label": "Choice One",
                        },
                        {
                            "name": "two",
                            "label": "Choice Two",
                        },
                        {
                            "name": "three",
                            "label": "Choice Three",
                        },
                    ],
                },
                {
                    "name": "grouped_choice",
                    "label": "Grouped Choice",
                    "hint": "Pick One",
                    "type": "select one",
                    "bind": {"required": True},
                    "choices": [
                        {
                            "name": "one",
                            "label": "Choice One",
                            "group": "Group 1",
                        },
                        {
                            "name": "two",
                            "label": "Choice Two",
                            "group": "Group 1",
                        },
                        {
                            "name": "three",
                            "label": "Choice Three",
                            "group": "Group 1",
                        },
                    ],
                },
            ],
            conf["form"],
        )

    def test_rest_config_json_boolean_choices(self):
        conf = self.get_config("itemtype")
        self.assertEqual(
            [
                {
                    "name": "name",
                    "label": "Name",
                    "type": "string",
                    "wq:length": 10,
                    "bind": {"required": True},
                },
                {
                    "name": "active",
                    "label": "Active",
                    "type": "select one",
                    "choices": [
                        {
                            "name": True,
                            "label": "Yes",
                        },
                        {
                            "name": False,
                            "label": "No",
                        },
                    ],
                },
            ],
            conf["form"],
        )

    def test_rest_config_json_rels(self):
        pconf = self.get_config("parent")
        self.assertEqual(
            {
                "name": "children",
                "label": "Children",
                "type": "repeat",
                "bind": {"required": True},
                "children": [
                    {
                        "name": "name",
                        "label": "Name",
                        "type": "string",
                        "wq:length": 10,
                        "bind": {"required": True},
                    }
                ],
            },
            self.get_field(pconf, "children"),
        )

        cconf = self.get_config("child")
        self.assertEqual(
            {
                "name": "parent",
                "label": "Parent",
                "type": "select one",
                "wq:ForeignKey": "parent",
                "wq:related_name": "children",
                "bind": {"required": True},
            },
            self.get_field(cconf, "parent"),
        )

    def test_rest_config_json_override(self):
        iconf = self.get_config("item")
        self.assertEqual(
            {
                "name": "type",
                "label": "Type",
                "type": "select one",
                "wq:ForeignKey": "itemtype",
                "wq:related_name": "item_set",
                "filter": {"active": ["1", "{{#id}}0{{/id}}{{^id}}1{{/id}}"]},
                "bind": {"required": True},
            },
            self.get_field(iconf, "type"),
        )

    def test_rest_config_json_label(self):
        pconf = self.get_config("slugrefparent")
        self.assertIn("label_template", pconf)
        self.assertEqual(
            "{{name}}{{#ref_id}} ({{ref}}){{/ref_id}}",
            pconf["label_template"],
        )

    @unittest.skipUnless(settings.WITH_GIS, "requires GIS")
    def test_rest_config_subtype_gis(self):
        conf = self.get_config("geometrymodel")
        field = self.get_field(conf, "geometry")
        self.assertEqual("geoshape", field["type"])

        conf = self.get_config("pointmodel")
        field = self.get_field(conf, "geometry")
        self.assertEqual("geopoint", field["type"])

    def test_rest_config_subtype(self):
        conf = self.get_config("filemodel")
        field = self.get_field(conf, "file")
        self.assertEqual("file", field["type"])

        conf = self.get_config("imagemodel")
        field = self.get_field(conf, "image")
        self.assertEqual("image", field["type"])

        conf = self.get_config("item")
        field = self.get_field(conf, "name")
        self.assertEqual("string", field["type"])

        conf = self.get_config("rootmodel")
        field = self.get_field(conf, "description")
        self.assertEqual("text", field["type"])

    def test_rest_config_field_order(self):
        conf = self.get_config("slugrefparent")
        self.assertEqual(conf["form"][0]["name"], "ref")
        self.assertEqual(conf["form"][1]["name"], "name")

    def test_rest_list_exclude_config(self):
        conf = self.get_config("expensivemodel")
        self.get_field(conf, "name")
        self.get_field(conf, "expensive")
        self.assertIsNone(
            self.get_field(conf, "more_expensive", allow_none=True)
        )

    def test_rest_virtual_fieldset(self):
        conf = self.get_config("fieldsetmodel")
        self.assertEqual(
            [
                {
                    "name": "general",
                    "type": "group",
                    "label": "General",
                    "children": [
                        {
                            "name": "name",
                            "label": "Name",
                            "type": "string",
                            "wq:length": 50,
                            "bind": {"required": True},
                        },
                        {
                            "name": "title",
                            "label": "Title",
                            "type": "string",
                            "wq:length": 20,
                            "bind": {"required": True},
                        },
                        {
                            "name": "status",
                            "label": "Status",
                            "type": "string",
                            "disabled": True,
                        },
                    ],
                },
                {
                    "name": "contact",
                    "type": "group",
                    "label": "Contact Information",
                    "control": {"appearance": "contact-fieldset"},
                    "children": [
                        {
                            "name": "address",
                            "label": "Address",
                            "type": "string",
                            "wq:length": 255,
                            "bind": {"required": True},
                        },
                        {
                            "name": "city",
                            "label": "City",
                            "type": "string",
                            "wq:length": 255,
                            "bind": {"required": True},
                        },
                    ],
                },
            ],
            conf["form"],
        )



from wq.db import rest
from .base import APITestCase
from rest_framework import status
from django.contrib.auth.models import User
from tests.patterns_app.models import Campaign, Entity, Attribute


class CustomPatternTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser", is_superuser=True)
        self.client.force_authenticate(user=self.user)
        self.campaign1 = Campaign.objects.create(
            slug="test-1",
            name="Test 1",
        )
        self.campaign2 = Campaign.objects.create(
            slug="test-2",
            name="Test 2",
        )
        self.att1 = Attribute.objects.create(
            name="Width",
            is_active=True,
            campaign=self.campaign1,
        )
        self.att2 = Attribute.objects.create(
            name="Height",
            is_active=False,
            campaign=self.campaign1,
        )
        self.att3 = Attribute.objects.create(
            name="Color",
            is_active=True,
            campaign=self.campaign2,
        )
        self.att4 = Attribute.objects.create(
            name="Size",
            is_active=False,
            campaign=self.campaign2,
        )
        self.entity1 = Entity.objects.create(
            name="Test", campaign=self.campaign1
        )

    def test_campaign_config(self):
        response = self.client.get("/config.json")
        self.assertEqual(
            [
                {
                    "name": "slug",
                    "label": "Slug",
                    "type": "string",
                    "bind": {"required": True},
                    "wq:length": 50,
                },
                {
                    "name": "name",
                    "label": "Name",
                    "type": "string",
                    "bind": {"required": True},
                    "wq:length": 10,
                },
                {
                    "name": "attributes",
                    "label": "Attributes",
                    "type": "repeat",
                    "bind": {"required": True},
                    "children": [
                        {
                            "name": "name",
                            "label": "Name",
                            "type": "string",
                            "bind": {"required": True},
                            "wq:length": 10,
                        },
                        {
                            "name": "is_active",
                            "label": "Is Active",
                            "type": "select one",
                            "choices": [
                                {"name": True, "label": "Yes"},
                                {"name": False, "label": "No"},
                            ],
                            "control": {"appearance": "checkbox"},
                        },
                    ],
                },
            ],
            response.data["pages"]["campaign"]["form"],
        )

    def test_eav_config(self):
        response = self.client.get("/config.json")
        self.assertEqual(
            [
                {
                    "name": "campaign",
                    "label": "Campaign",
                    "type": "select one",
                    "wq:ForeignKey": "campaign",
                    "wq:related_name": "entity_set",
                    "bind": {"required": True},
                    "control": {"appearance": "campaign-label"},
                },
                {
                    "name": "general",
                    "label": "Info",
                    "type": "group",
                    "children": [
                        {
                            "name": "name",
                            "label": "Name",
                            "type": "string",
                            "bind": {"required": True},
                            "wq:length": 10,
                        },
                    ],
                    "control": {"appearance": "fieldset"},
                },
                {
                    "name": "values",
                    "label": "Values",
                    "type": "repeat",
                    "bind": {"required": True},
                    "children": [
                        {
                            "name": "attribute",
                            "label": "Attribute",
                            "type": "select one",
                            "wq:ForeignKey": "attribute",
                            "wq:related_name": "value_set",
                            "bind": {"required": True},
                            "control": {"appearance": "attribute-label"},
                        },
                        {
                            "name": "value",
                            "label": "Value",
                            "type": "text",
                            "bind": {"required": True},
                        },
                    ],
                    "control": {"appearance": "eav-fieldset-array"},
                    "initial": {
                        "type_field": "attribute",
                        "filter": {
                            "campaign": "{{campaign_id}}",
                            "is_active": True,
                        },
                    },
                },
            ],
            response.data["pages"]["entity"]["form"],
        )

    def test_campaign_post(self):
        form = {
            "slug": "test-3",
            "name": "Test 3",
            "attributes[0][name]": "Start Temp",
            "attributes[1][name]": "End Temp",
        }

        response = self.client.post("/campaigns.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data["name"], "Test 3")
        self.assertIn("attributes", response.data)
        atts = response.data["attributes"]
        self.assertEqual(len(atts), 2)
        self.assertEqual(atts[0]["name"], "Start Temp")

    def test_campaign_put(self):
        form = {
            "slug": self.campaign1.slug,
            "name": "Test 1'",
            "attributes[0][name]": "Width (ft)",
        }

        response = self.client.put(
            "/campaigns/%s.json" % self.campaign1.slug, form
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.campaign1.refresh_from_db()
        self.assertEqual(self.campaign1.name, "Test 1'")
        self.assertEqual(self.campaign1.attributes.count(), 1)
        self.assertEqual(self.campaign1.attributes.first().name, "Width (ft)")

    def test_eav_post(self):
        form = {
            "campaign_id": self.campaign1.slug,
            "general[name]": "Test A",
            "values[0][attribute_id]": self.att1.pk,
            "values[0][value]": "100",
        }

        response = self.client.post("/entities.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data["general"]["name"], "Test A")
        self.assertIn("values", response.data)
        atts = response.data["values"]
        self.assertEqual(len(atts), 1)
        self.assertEqual(atts[0]["attribute_id"], self.att1.pk)

    def test_eav_post_empty(self):
        form = {
            "campaign_id": self.campaign1.slug,
            "general": {
                "name": "Test B",
            },
            "values": [
                {"value": "", "attribute_id": self.att1.pk},
                {"value": 0, "attribute_id": self.att2.pk},
            ],
        }

        response = self.client.post("/entities.json", form, format="json")
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertIn("values", response.data)
        atts = response.data["values"]
        self.assertEqual(len(atts), 1)
        self.assertEqual(atts[0]["attribute_id"], self.att2.pk)
        self.assertEqual(atts[0]["value"], "0")

    def test_eav_put(self):
        form = {
            "name": "Test'",
            "campaign_id": self.campaign2.slug,
            "values[0][value]": "Red",
            "values[0][attribute_id]": self.att3.pk,
        }

        response = self.client.put("/entities/%s.json" % self.entity1.pk, form)
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.entity1.refresh_from_db()
        self.assertEqual(self.entity1.name, "Test'")
        self.assertEqual(self.entity1.values.count(), 1)
        self.assertEqual(self.entity1.values.first().attribute, self.att3)



from rest_framework.generics import GenericAPIView as RestGenericAPIView
from rest_framework.response import Response
from rest_framework.decorators import action
from rest_framework import status, viewsets
from django.db.models import ProtectedError


class GenericAPIView(RestGenericAPIView):
    router = None
    ignore_kwargs = []

    @property
    def template_name(self):
        """
        Infer template name from view/viewset name
        """
        name = type(self).__name__
        name = name.replace("ViewSet", "")
        name = name.replace("View", "")
        return name.lower() + ".html"

    @property
    def depth(self):
        return 0

    def get_template_names(self):
        return [self.template_name]

    def get_queryset(self):
        if self.router is not None and self.model is not None:
            return self.router.get_queryset_for_model(self.model, self.request)
        return super(GenericAPIView, self).get_queryset()

    def get_serializer_class(self):
        if self.router is not None and self.model is not None:
            return self.router.get_serializer_for_model(self.model, self.depth)
        return super(GenericAPIView, self).get_serializer_class()

    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class SimpleView(GenericAPIView):
    def get(self, request, *args, **kwargs):
        return Response({})


class SimpleViewSet(viewsets.ViewSet, GenericAPIView):
    def list(self, request, *args, **kwargs):
        return Response({})


class ModelViewSet(viewsets.ModelViewSet, GenericAPIView):
    @property
    def template_name(self):
        basename = self.model._meta.model_name
        if self.action in ("retrieve", "create", "update", "delete"):
            suffix = "detail"
        elif self.action and self.action != "metadata":
            suffix = self.action
        else:
            suffix = "list"
        return "%s_%s.html" % (basename, suffix)

    @property
    def depth(self):
        if self.action in ("retrieve", "edit"):
            return 1
        else:
            return 0

    @action(detail=True)
    def edit(self, request, *args, **kwargs):
        """
        Edit view (note: form and lookup context now generated on client)
        """
        return self.retrieve(request, *args, **kwargs)

    def new(self, request):
        """
        new is a variant of the "edit" action, but with no existing model
        to lookup.
        """
        self.action = "edit"
        return Response({})

    def retrieve(self, request, *args, **kwargs):
        """
        Custom retrieve watches for "new" lookup value and switches modes
        accordingly
        """
        lookup = self.lookup_url_kwarg or self.lookup_field

        if self.kwargs.get(lookup, "") == "new":
            # new/edit mode
            return self.new(request)
        else:
            # Normal detail view
            return super(ModelViewSet, self).retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).list(request, *args, **kwargs)
        if not isinstance(response.data, dict):
            return response

        if not self.router:
            return response

        if len(set(self.kwargs) - {"format"}) == 0:
            return response

        return self.add_parent_info(request, response)

    def create(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).create(request, *args, **kwargs)
        if not request.accepted_media_type.startswith("text/html"):
            # JSON request, assume client will handle redirect
            return response

        # HTML request, probably a form post from an older browser
        if response.status_code == status.HTTP_201_CREATED:
            return self.postsave(request, response)
        else:
            return self.saveerror(request, response)

    def update(self, request, *args, **kwargs):
        response = super(ModelViewSet, self).update(request, *args, **kwargs)
        if not request.accepted_media_type.startswith("text/html"):
            # JSON request, assume client will handle redirect
            return response

        # HTML request, probably a form post from an older browser
        if response.status_code == status.HTTP_200_OK:
            return self.postsave(request, response)
        else:
            return self.saveerror(request, response)

    def postsave(self, request, response):
        if self.router:
            conf = self.router.get_config_for_model(self.model)
        else:
            conf = {}

        # Redirect to new page
        postsave = conf.get(
            "postsave", self.model._meta.model_name + "_detail"
        )
        if "_" in postsave:
            page, mode = postsave.split("_")
        else:
            page = postsave
            mode = "detail"

        oid = ""
        if page != self.model._meta.model_name and self.router:
            # Optional: return to detail view of a parent model
            pconf = self.router.get_page_config(page)
            if pconf.get("list", None) and mode != "list":
                oid = response.data.get(page + "_id", None)
        else:
            # Default: return to detail view of the saved model
            pconf = conf
            if mode != "list":
                oid = response.data["id"]

        url = "/" + pconf["url"]
        if pconf["url"] and pconf.get("list", None):
            url += "/"
        if oid:
            url += str(oid)
            if mode == "edit":
                url += "/edit"

        return Response(
            {"detail": "Created"},
            status=status.HTTP_302_FOUND,
            headers={"Location": url},
        )

    def saveerror(self, request, response):
        errors = [
            {"field": key, "errors": val} for key, val in response.data.items()
        ]
        template = self.model._meta.model_name + "_error.html"
        return Response(
            {"errors": errors, "post": request.DATA},
            status=response.status_code,
            template_name=template,
        )

    def destroy(self, request, *args, **kwargs):
        try:
            response = super(ModelViewSet, self).destroy(
                request, *args, **kwargs
            )
        except ProtectedError as e:
            return Response({"non_field_errors": [e.args[0]]}, status=400)
        else:
            return response

    def add_parent_info(self, request, response):
        parent_model = None
        pid = None
        for rel_model, fields in self.router.get_foreign_keys(
            self.model
        ).items():
            if len(fields) == 1 and fields[0] in self.kwargs:
                pid = self.kwargs[fields[0]]
                parent_model = rel_model

        if not parent_model:
            return response

        parent = self.router.get_by_identifier(
            self.router.get_queryset_for_model(parent_model, request), pid
        )
        objid = self.router.get_object_id(parent)
        page_config = self.router.get_model_config(parent_model)
        urlbase = page_config["url"]
        if urlbase != "":
            urlbase += "/"
        response.data["parent_label"] = str(parent)
        response.data["parent_id"] = objid
        response.data["parent_url"] = "%s%s" % (urlbase, objid)
        response.data["parent_is_" + page_config["name"]] = True
        response.data["parent_page"] = page_config["name"]
        response.data["page_config"] = page_config
        response.data["parent"] = self.router.serialize(
            parent,
            request=self.request,
        )
        return response



from wq.db import rest
from .models import GeometryModel, PointModel
from django.conf import settings

if settings.WITH_GIS:
    rest.router.register(GeometryModel, fields="__all__")
    rest.router.register(PointModel, fields="__all__")



from .base import APITestCase
import json

from django.contrib.auth.models import User


class AuthTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser")
        self.client.force_authenticate(self.user)

    def test_auth_config_json(self):
        response = self.client.get("/config.json")
        result = json.loads(response.content.decode("utf-8"))
        self.assertTrue("login" in result["pages"])

    def test_auth_login_info(self):
        response = self.client.get("/login.json")
        result = json.loads(response.content.decode("utf-8"))
        self.assertTrue("user" in result)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-31 14:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('conflict_app', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
    ]



from wq.db import rest
from .views import LoginView, LogoutView
from .serializers import UserSerializer
from django.contrib.auth.models import User

rest.router.add_page("login", {"url": "login"}, LoginView)
rest.router.add_page("logout", {"url": "logout"}, LogoutView)

# Configure UserSerializer (without actually registering User model)
rest.router.register_serializer(User, UserSerializer)



import os


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": (
                "django.template.context_processors.csrf",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            )
        },
    }
]

REST_FRAMEWORK = {
    "DEFAULT_RENDERER_CLASSES": (
        "wq.db.rest.renderers.HTMLRenderer",
        "wq.db.rest.renderers.JSONRenderer",
        "wq.db.rest.renderers.GeoJSONRenderer",
    ),
    "DEFAULT_PAGINATION_CLASS": "wq.db.rest.pagination.Pagination",
    "PAGE_SIZE": 50,
    "DEFAULT_PERMISSION_CLASSES": ("wq.db.rest.permissions.ModelPermissions",),
    "DEFAULT_FILTER_BACKENDS": ("wq.db.rest.filters.FilterBackend",),
    "DEFAULT_CONTENT_NEGOTIATION_CLASS": "wq.db.rest.negotiation.ContentNegotiation",
}

SECRET_KEY = "1234"

MIDDLEWARE_CLASSES = ("django.middleware.csrf.CsrfViewMiddleware",)

INSTALLED_APPS = (
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "wq.db.rest",
    "wq.db.rest.auth",
    "tests.rest_app",
    "tests.conflict_app",
    "tests.patterns_app",
    "tests.naturalkey_app",
)

VARIANT = os.environ.get("TEST_VARIANT")

if VARIANT in ("postgis", "spatialite"):
    WITH_GIS = True
else:
    WITH_GIS = False

if VARIANT == "nonroot":
    WITH_NONROOT = True
else:
    WITH_NONROOT = False

if WITH_GIS:
    INSTALLED_APPS += ("tests.gis_app",)


if VARIANT in ("postgis", "postgres"):
    if WITH_GIS:
        engine = "django.contrib.gis.db.backends.postgis"
    else:
        engine = "django.db.backends.postgresql"
    DATABASES = {
        "default": {
            "ENGINE": engine,
            "NAME": "wqdb_test",
            "USER": os.environ.get("USER"),
        }
    }
else:
    if WITH_GIS:
        engine = "django.contrib.gis.db.backends.spatialite"
        SPATIALITE_LIBRARY_PATH = "mod_spatialite.so"
    else:
        engine = "django.db.backends.sqlite3"
    DATABASES = {
        "default": {
            "ENGINE": engine,
            "NAME": "wqdb_test.sqlite3",
        }
    }

USE_TZ = True
TIME_ZONE = "America/Chicago"

ROOT_URLCONF = "tests.urls"
BASE_DIR = os.path.dirname(__file__)
MEDIA_ROOT = os.path.join(BASE_DIR, "media")
VERSION_TXT = os.path.join(BASE_DIR, "version.txt")
WQ_APP_TEMPLATE = os.path.join(BASE_DIR, "index.html")

DEBUG = True



from .base import APITestCase
from rest_framework import status
from django.contrib.auth.models import User
import json
from tests.naturalkey_app.models import ModelWithNaturalKey


class NaturalKeyTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser", is_superuser=True)
        self.client.force_authenticate(user=self.user)

    def test_naturalkey_config(self):
        response = self.client.get("/config.json")
        self.assertEqual(
            [
                {
                    "name": "note",
                    "label": "Note",
                    "type": "text",
                    "bind": {"required": True},
                },
                {
                    "name": "key",
                    "label": "Key",
                    "type": "group",
                    "bind": {"required": True},
                    "children": [
                        {
                            "name": "parent[slug]",
                            "label": "Parent",
                            "type": "select one",
                            "wq:ForeignKey": "naturalkeyparent",
                            "wq:length": 50,
                            "bind": {"required": True},
                        },
                        {
                            "name": "date",
                            "label": "Date",
                            "type": "date",
                            "bind": {"required": True},
                        },
                    ],
                },
            ],
            response.data["pages"]["modelwithnaturalkey"]["form"],
        )

    def test_naturalkey_post(self):
        form = {
            "key[parent][slug]": "test-key",
            "key[date]": "2016-12-31",
            "note": "Test Note",
        }

        response = self.client.post("/modelwithnaturalkeys.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        instance = ModelWithNaturalKey.objects.get(pk=response.data["id"])

        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {
                "id": instance.id,
                "note": "Test Note",
                "key": {
                    "parent": {"slug": "test-key"},
                    "date": "2016-12-31",
                },
                "label": "test-key on 2016-12-31: Test Note",
            },
        )



# Generated by Django 4.1.5 on 2023-01-11 04:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Attribute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=10)),
                ("is_active", models.BooleanField(default=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Campaign",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("slug", models.SlugField()),
                ("name", models.CharField(max_length=10)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Entity",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=10)),
                (
                    "campaign",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="patterns_app.campaign",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "entities",
            },
        ),
        migrations.CreateModel(
            name="Value",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.TextField()),
                (
                    "attribute",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="patterns_app.attribute",
                    ),
                ),
                (
                    "entity",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="values",
                        to="patterns_app.entity",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="attribute",
            name="campaign",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="attributes",
                to="patterns_app.campaign",
            ),
        ),
    ]



from django.urls import path
from wq.db import rest
from .settings import WITH_NONROOT


if WITH_NONROOT:
    base_url = "wqsite/"
else:
    base_url = ""


urlpatterns = [
    path(base_url, rest.router.urls),
]



from .base import APITestCase
from django.core.exceptions import ImproperlyConfigured


class RestRouterTestCase(APITestCase):
    def test_rest_model_conflict(self):
        from wq.db import rest
        from tests.conflict_app.models import Item

        # Register model with same name as existing model
        with self.assertRaises(ImproperlyConfigured) as e:
            rest.router.register_model(Item, fields="__all__")
        self.assertEqual(
            e.exception.args[0],
            "Could not register <class 'tests.conflict_app.models.Item'>: "
            "the name 'item' was already registered for "
            "<class 'tests.rest_app.models.Item'>",
        )
        self.assertNotIn(Item, rest.router._models)

        # Register model with different name, but same URL as existing model
        with self.assertRaises(ImproperlyConfigured) as e:
            rest.router.register_model(
                Item, name="conflictitem", fields="__all__"
            )
        self.assertEqual(
            e.exception.args[0],
            "Could not register <class 'tests.conflict_app.models.Item'>: "
            "the url 'items' was already registered for "
            "<class 'tests.rest_app.models.Item'>",
        )
        self.assertNotIn(Item, rest.router._models)

        # Register model with different name and URL
        rest.router.register_model(
            Item, name="conflictitem", url="conflictitems", fields="__all__"
        )
        self.assertIn(Item, rest.router._models)
        self.assertIn("conflictitem", rest.router.config["pages"])

    def test_rest_old_config(self):
        from wq.db import rest
        from tests.conflict_app.models import TestModel

        with self.assertRaises(ImproperlyConfigured):
            rest.router.register_model(
                TestModel, partial=True, fields="__all__"
            )

        self.assertNotIn(TestModel, rest.router._models)

        with self.assertRaises(ImproperlyConfigured):
            rest.router.register_model(
                TestModel, reversed=True, fields="__all__"
            )

        self.assertNotIn(TestModel, rest.router._models)

        with self.assertRaises(ImproperlyConfigured):
            rest.router.register_model(
                TestModel, max_local_pages=0, fields="__all__"
            )

        self.assertNotIn(TestModel, rest.router._models)
